use crate::Peer;
use number_prefix::{NumberPrefix, Prefixed, Standalone};
use std::env;
use std::net::SocketAddr;
use std::path::PathBuf;
use std::process::Command;
use std::thread;
use std::time::SystemTime;
use threadpool::ThreadPool;

//#[non_exhaustive]
#[derive(Debug, Clone)]
pub enum Event<'a> {
    Added(&'a Peer),
    Removed(&'a Peer),
    Connected(&'a Peer),
    Disconnected(&'a Peer),
    Roamed(&'a Peer, SocketAddr),
    Transfered { peer: &'a Peer, tx: u64, rx: u64 },
}

pub trait EventListener {
    fn name(&self) -> &'static str;

    fn fire<'a>(&self, event: Event<'a>);
}

impl EventListener for Vec<Box<EventListener>> {
    fn name(&self) -> &'static str {
        "List of Listeners"
    }
    fn fire<'a>(&self, event: Event<'a>) {
        self.iter().for_each(|l| l.fire(event.clone()));
    }
}

pub struct LogListener {
    log_transfer: bool,
}

impl Default for LogListener {
    fn default() -> LogListener {
        let log = env::var("WG_LOG_TRANSFER")
            .map(|t| t.parse::<bool>().unwrap_or(true))
            .unwrap_or(false);
        LogListener { log_transfer: log }
    }
}

impl EventListener for LogListener {
    fn name(&self) -> &'static str {
        "Log"
    }

    fn fire<'a>(&self, event: Event<'a>) {
        match event {
            Event::Connected(peer) => println!("{} connected!", peer.key),
            Event::Disconnected(peer) => println!("{} disconnected!", peer.key),
            Event::Added(peer) => println!("{} added!", peer.key),
            Event::Removed(peer) => println!("{} removed!", peer.key),
            Event::Roamed(peer, previous_addr) => println!(
                "{} roamed {} -> {}!",
                peer.key,
                previous_addr,
                peer.endpoint.unwrap()
            ),
            Event::Transfered { peer, tx, rx } if self.log_transfer => {
                let prefix = |cnt: u64| match NumberPrefix::binary(cnt as f64) {
                    Standalone(bytes) => format!("{}B", bytes),
                    Prefixed(prefix, n) => format!("{:.0} {}B", n, prefix),
                };
                println!(
                    "{} transfered {} up, {} down",
                    peer.key,
                    prefix(tx),
                    prefix(rx)
                );
            }
            Event::Transfered { peer, tx, rx } => (),
            _ => unimplemented!(),
        }
    }
}

pub struct ScriptListener {
    pub script: PathBuf,
    pub pool: ThreadPool,
}

impl ScriptListener {
    pub fn new(script: PathBuf) -> ScriptListener {
        let pool = ThreadPool::new(8);
        ScriptListener { script, pool }
    }

    fn peer_props<'a>(&self, peer: &'a Peer) -> String {
        format!(
            "{id} {allowed_ips} {endpoint} {last_handshake} {persistent_keepalive} {traffic}",
            id = peer.key,
            allowed_ips = peer
                .allowed_ips
                .iter()
                .map(|(addr, mask)| [addr.to_string(), mask.to_string()].join("/"))
                .collect::<Vec<String>>()
                .join(","),
            endpoint = peer
                .endpoint
                .map(|e| e.to_string())
                .unwrap_or("0".to_owned()),
            last_handshake = peer
                .last_handshake
                .map(|s| s.duration_since(SystemTime::UNIX_EPOCH).unwrap().as_secs() as i64)
                .unwrap_or(-1),
            persistent_keepalive = peer
                .persistent_keepalive
                .map(|k| k.as_secs() as i64)
                .unwrap_or(-1),
            traffic = {
                let (rx, tx) = peer.traffic;
                [rx.to_string(), tx.to_string()].join(",")
            }
        )
    }

    fn mkcmd<'a>(&self, args: Vec<&'a str>) -> Command {
        let mut cmd = Command::new("/bin/sh");
        cmd.arg("-c");
        cmd.arg(format!(
            "{} {}",
            (&self.script).to_str().unwrap(),
            args.join(" ")
        ));
        cmd
    }

    fn call_sub<'a>(&self, args: Vec<&'a str>) {
        let mut cmd = self.mkcmd(args);
        self.pool.execute(move || {
            cmd.output().expect("Failed to call Script hooḱ!");
        });
    }
}

impl EventListener for ScriptListener {
    fn name(&self) -> &'static str {
        "Script"
    }
    fn fire<'a>(&self, event: Event<'a>) {
        match event {
            Event::Connected(peer) => self.call_sub(vec!["connected", &self.peer_props(peer)]),
            Event::Disconnected(peer) => {
                self.call_sub(vec!["disconnected", &self.peer_props(peer)])
            }
            Event::Added(peer) => self.call_sub(vec!["added", &self.peer_props(peer)]),
            Event::Removed(peer) => self.call_sub(vec!["removed", &self.peer_props(peer)]),
            Event::Roamed(peer, previous_addr) => self.call_sub(vec![
                "roaming",
                &self.peer_props(peer),
                &previous_addr.to_string(),
            ]),
            Event::Transfered { peer, tx, rx } => self.call_sub(vec![
                "transfered",
                &self.peer_props(peer),
                &tx.to_string(),
                &rx.to_string(),
            ]),
            _ => unimplemented!(),
        }
    }
}
